using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;

namespace sampleIME.control
{
    /// <summary>
    /// Cxgp̈
    /// </summary>
    public class StringEventArgs : EventArgs
    {
        // 
        private string text;

        public StringEventArgs(string text)
        {
            this.text = text;
        }

        public string Text
        {
            get { return text; }
        }
    }

    public class ConversionEventArgs : EventArgs
    {
        // ϊ
        private string[] candidate;

        // I
        private int select;

        // ϊ⑍
        private int countAll;

        public ConversionEventArgs(string[] candidate, int select, int countAll)
        {
            this.candidate = candidate;
            this.select = select;
            this.countAll = countAll;
        }

        public string[] Candidate
        {
            get { return candidate; }
        }

        public int Selection
        {
            get { return select; }
        }

        public int CountAll
        {
            get { return countAll; }
        }
    }

    /// <summary>
    /// L[Cxgp̈
    /// </summary>
    public class KeyEventArgs : EventArgs
	{
        // L[
		private Keys keyCode;

		public KeyEventArgs( Keys keyCode )
		{
			this.keyCode = keyCode;
		}

		public Keys KeyCode
		{
			get { return keyCode; }
		}
	}

    // Cxgnh
    public delegate void InputEventHandler(object sender, StringEventArgs e);

    public delegate void ConvertEventHandler(object sender, ConversionEventArgs e);

    public delegate void KeyEventHandler(object sender, KeyEventArgs e);

	public static class TextInput
	{
		/// <summary>
		/// ̓Cxg
		/// </summary>
        public static event InputEventHandler InputtedText;

        /// <summary>
        /// ̓Cxg
        /// </summary>
        public static event InputEventHandler InputtingText;

        /// <summary>
        /// ϊCxg
        /// </summary>
        public static event ConvertEventHandler ChangeCandidate;

        /// <summary>
		/// L[_ECxg
		/// </summary>
		public static event KeyEventHandler KeyDown;

		/// <summary>
		/// L[[XCxg
		/// </summary>
		public static event KeyEventHandler KeyUp;

        delegate IntPtr WndProc(IntPtr hWnd, uint msg, IntPtr wParam, IntPtr lParam);

		static bool initialized;
		static IntPtr prevProc;
		static WndProc hookProc;
        static Ime ime;
        static bool enabled;

		const int GWL_WNDPROC = -4;
        const int DLGC_WANTALLKEYS = 4;

        // Window bZ[W
        const int WM_GETDLGCODE = 0x0087;
        const int WM_KEYDOWN = 0x0100;
        const int WM_KEYUP = 0x0101;
        const int WM_CHAR = 0x0102;
        const int WM_IME_STARTCOMPOSITION = 0x010D;
        const int WM_IME_ENDCOMPOSITION = 0x010E;
        const int WM_IME_COMPOSITION = 0x010F;
        const int WM_IME_SETCONTEXT = 0x0281;
        const int WM_IME_NOTIFY = 0x0282;

        // ϊEBhE
        const int IMN_CHANGECANDIDATE = 0x0003;
        const int IMN_OPENCANDIDATE = 0x0005;
        const int IMN_SETOPENSTATUS = 0x0008;

		[DllImport( "user32.dll" )]
		static extern IntPtr CallWindowProc( IntPtr lpPrevWndFunc, IntPtr hWnd, uint Msg, IntPtr wParam, IntPtr lParam );

		[DllImport( "user32.dll" )]
		static extern int SetWindowLong( IntPtr hWnd, int nIndex, int dwNewLong );
        
        /// <summary>
		/// 
		/// </summary>
		/// <param name="window"></param>
		public static void Initialize( GameWindow window )
		{
            if (initialized) return;

            hookProc = new WndProc(HookProc);
            prevProc = (IntPtr)SetWindowLong(window.Handle, GWL_WNDPROC,
                (int)Marshal.GetFunctionPointerForDelegate(hookProc));
            ime = new Ime(window.Handle);
            initialized = true;
        }

        /// <summary>
        /// bZ[WtbN
        /// </summary>
		static IntPtr HookProc(IntPtr hWnd, uint msg, IntPtr wParam, IntPtr lParam)
		{
			IntPtr returnCode = CallWindowProc(prevProc, hWnd, msg, wParam, lParam);
            int wp = wParam.ToInt32();
            int lp = lParam.ToInt32();

			switch(msg)
			{
                case WM_GETDLGCODE:
                    returnCode = (IntPtr)(returnCode.ToInt32() | DLGC_WANTALLKEYS);
                    break;

				case WM_KEYDOWN:
                    if (KeyDown != null) 
                        KeyDown(null, new KeyEventArgs((Keys)wParam));
					break;

				case WM_KEYUP:
					if (KeyUp != null) 
                        KeyUp(null, new KeyEventArgs((Keys)wParam));
					break;

				case WM_CHAR:
                    InputTextProcess(wp, lp);
                    break;

				case WM_IME_SETCONTEXT:
                    if (wp == 1)
                        ime.AssociateContext();
                    break;

                case WM_IME_STARTCOMPOSITION:
                    ime.Initalize();
                    break;

                case WM_IME_ENDCOMPOSITION:
                    // ͊mCxg
                    InputtedText(null, new StringEventArgs(ime.Inputted()));
                    break;

                case WM_IME_COMPOSITION:
                    // ͒Cxg
                    InputtingText(null, new StringEventArgs(ime.Inputting()));
                    break;

                case WM_IME_NOTIFY:

                    if (wp == IMN_SETOPENSTATUS) {
                        enabled = !enabled;
                    } else if (wp == IMN_OPENCANDIDATE || wp == IMN_CHANGECANDIDATE) {
                        // ϊύXCxg
                        if (ChangeCandidate != null)
                            ChangeCandidate(null, ime.ConversionCandidate());
                    }
                    break;
            }
			return returnCode;
		}

        /// <summary>
        /// ͏
        /// </summary>
        /// <param name="wp"></param>
        /// <param name="lp"></param>
        static void InputTextProcess(int wp, int lp)
        {
            if (InputtedText == null || lp == 1) return;

            char c = (char)wp;
            InputTextProcess(c.ToString());
        }

        /// <summary>
        /// ͏
        /// </summary>
        /// <param name="text"></param>
        static void InputTextProcess(string text)
        {
            if (InputtedText == null) return;

            InputtedText(null, new StringEventArgs(text));
        }

        /// <summary>
        /// IME ON/OFF
        /// </summary>
        public static bool Enabled
        {
            get { return enabled; }
        }
    }
}

